/* ARCADEmini USB driver for Linux test application
 * version 1.2.0 date 2005-06-06
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/poll.h>

#include "am_usb.h"



#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))



char * SNES_KEYS[16] = { "?0", "?1", "?2", "?3", "R", "L", "X", "A", "Right", "Left", "Down", "Up", "Start", "Select", "Y", "B" };



//get first word and rest of a string
static void str_first_word( char * p_string, char * * pp_first, char * * pp_rest )
{
  //jump over whitespaces
  while( *p_string == '\t' || *p_string == ' ' )
    p_string++;

  //first word starts here
  *pp_first = p_string;

  //jump over first word
  while( *p_string != 0 && *p_string != '\t' && *p_string != ' ' )
    p_string++;

  //first word ends here
  if( *p_string != 0 )
  {
    *p_string = 0;
    p_string++;
  }

  //jump over whitespaces
  while( *p_string == '\t' || *p_string == ' ' )
    p_string++;

  //rest starts here
  *pp_rest = p_string;
}

//check if commands are executed from compact flash or internal flash
static void cmd_cf_flag( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_cf_flag get_cf_flag;
  int err;
  
  //get compact flash flag
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_cf_flag, &get_cf_flag );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show if commands are executed from compact flash or internal flash
  else
    if( get_cf_flag.cf_flag == 0 )
      printf( "  commands are executed from internal flash\n" );
    else
      printf( "  commands are executed from compact flash\n" );
}

//read compact flash sector into sector buffer
static void cmd_cf_read( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  struct t_am_usb_read_cf read_cf;
  int err;

  //get sector number
  str_first_word( p_params, &p_param, &p_params );
  read_cf.sec_no = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 )
  {
    printf( "  invalid sector number: \"%s\"\n", p_param );
    return;
  }
     
  //read sector of compact flash
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_cf, &read_cf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector number and result
  else
    if( read_cf.err == 0 )
      printf( "  compact flash sector %ld successfully read\n", read_cf.sec_no );
    else
      printf( "  reading compact flash sector %ld failed\n"
              "   - no CF inserted?\n"
              "   - reading past end of CF?\n", read_cf.sec_no );
}

//get number of sectors on compact flash
static void cmd_cf_secs( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_cf_sec_cnt get_cf_sec_cnt;
  int err;

  //get number of sectors on compact flash
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_cf_sec_cnt, &get_cf_sec_cnt );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show number of sectors and size
  else
    printf( "  compact flash has got %ld sectors (%f MB)\n",
            get_cf_sec_cnt.sec_cnt,
            (double)get_cf_sec_cnt.sec_cnt / 2048.0 );
}

//write sector buffer to compact flash sector
static void cmd_cf_write( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  struct t_am_usb_write_cf write_cf;
  int err;

  //get sector number
  str_first_word( p_params, &p_param, &p_params );
  write_cf.sec_no = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 )
  {
    printf( "  invalid sector number: \"%s\"\n", p_param );
    return;
  }
     
  //write sector of compact flash
  err = ioctl( fd_am_usb, am_usb_ioctl_no_write_cf, &write_cf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector number and result
  else
    if( write_cf.err == 0 )
      printf( "  compact flash sector %ld successfully written\n", write_cf.sec_no );
    else
      printf( "  writing compact flash sector %ld failed\n"
              "   - no CF inserted?\n"
              "   - writing past end of CF?\n", write_cf.sec_no );
}

//get command pointer
static void cmd_cmd_ptr_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_cmd_ptr get_cmd_ptr;
  int err;

  //get command pointer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_cmd_ptr, &get_cmd_ptr );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show command pointer
  else
    printf( "  command pointer: sector %ld offset %X\n",
            get_cmd_ptr.sec_no, get_cmd_ptr.ofs );
}

//set command pointer
static void cmd_cmd_ptr_set( int fd_am_usb, char * p_params )
{
  struct t_am_usb_set_cmd_ptr set_cmd_ptr;
  char * p_param, * p_chr;
  unsigned long val;
  int err;

  //get sector number
  str_first_word( p_params, &p_param, &p_params );
  val = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 )
  {
    printf( "  invalid sector number: \"%s\"\n", p_param );
    return;
  }
  set_cmd_ptr.sec_no = val;

  //get offset
  str_first_word( p_params, &p_param, &p_params );
  val = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || val > 511 )
  {
    printf( "  invalid offset: \"%s\"\n", p_param );
    return;
  }
  set_cmd_ptr.ofs = (unsigned short)val;

  //set command pointer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_cmd_ptr, &set_cmd_ptr );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show command pointer
  else
    printf( "  command pointer: sector %ld offset %X\n",
            set_cmd_ptr.sec_no, set_cmd_ptr.ofs );
}

//get version of firmware
static void cmd_fw_ver( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_fw_version get_fw_version;
  int err;

  //get firmware version
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_fw_version, &get_fw_version );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show firmware version
  else
    printf( "  firmware version %d.%d.%d date %04d-%02d-%02d\n",
            get_fw_version.major, get_fw_version.minor, get_fw_version.revision,
            2000 + get_fw_version.year, get_fw_version.month, get_fw_version.day );
}

//halt ARCADEmini
static void cmd_halt( int fd_am_usb, char * p_params )
{
  int err;

  //halt ARCADEmini
  err = ioctl( fd_am_usb, am_usb_ioctl_no_halt, NULL );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
}

//get ID of this ARCADEmini
static void cmd_id_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_id get_id;
  int err, i;

  //get identification
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_id, &get_id );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show identification
  else
  {
    //success
    if( get_id.err == 0 )
    {
      printf( "  identification: " );
      for( i = 0; i < sizeof( get_id.id ); i++ )
        if( (char)get_id.id[i] >= ' ' )
          printf( "%c", (char)get_id.id[i] );
        else
          printf( " " );
      printf( "\n" );
    }
    //error
    else
      printf( "  reading identification failed (maybe EEPROM is busy)\n" );
  }
}

//set ID of this ARCADEmini
static void cmd_id_set( int fd_am_usb, char * p_params )
{
  char * p_param;
  struct t_am_usb_set_id set_id;
  int i, err;

  //get identification from parameter
  str_first_word( p_params, &p_param, &p_params );
  for( i = 0; p_param[i] != 0 && i < sizeof( set_id.id ); i++ )
    set_id.id[i] = (unsigned char)p_param[i];
  for( ; i < sizeof( set_id.id ); i++ )
    set_id.id[i] = (unsigned char)' ';

  //set identification
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_id, &set_id );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show result
  else
  {
    //error
    if( set_id.err != 0 )
      printf( "  writing identification failed (maybe EEPROM is busy)\n" );
  }
}

//read internal flash sector into sector buffer
static void cmd_if_read( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  struct t_am_usb_read_if read_if;
  int err;

  //get sector number
  str_first_word( p_params, &p_param, &p_params );
  read_if.sec_no = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 )
  {
    printf( "  invalid sector number: \"%s\"\n", p_param );
    return;
  }
     
  //read sector of internal flash
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_if, &read_if );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector number and result
  else
    if( read_if.err == 0 )
      printf( "  internal flash sector %ld successfully read\n", read_if.sec_no );
    else
      printf( "  reading internal flash sector %ld failed\n"
              "   - reading past end of internal flash?\n", read_if.sec_no );
}

//get number of sectors in internal flash
static void cmd_if_secs( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_if_sec_cnt get_if_sec_cnt;
  int err;

  //get number of sectors in internal flash
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_if_sec_cnt, &get_if_sec_cnt );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show number of sectors and size
  else
    printf( "  internal flash has got %ld sectors (%f KB)\n",
            get_if_sec_cnt.sec_cnt,
            (double)get_if_sec_cnt.sec_cnt / 2 );
}

//get key state (and clear it)
static void cmd_keys( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_set_keys get_set_keys;
  int err, i;
  unsigned char mask;

  //clear key state
  get_set_keys.in_pressed = 0x00;
  get_set_keys.in_pressed_mask = 0xFF;
  get_set_keys.in_released = 0x00;
  get_set_keys.in_released_mask = 0xFF;

  //get and set key state
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_set_keys, &get_set_keys );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show key state
  else
  {
    printf( "  state:" );
    for( i = 0, mask = 1; i < 8; i++, mask <<= 1 )
      if( get_set_keys.state & mask )
        printf( " %d", i + 1 );
    printf( "\n" );
    printf( "  pressed:" );
    for( i = 0, mask = 1; i < 8; i++, mask <<= 1 )
      if( get_set_keys.pressed & mask )
        printf( " %d", i + 1 );
    printf( "\n" );
    printf( "  released:" );
    for( i = 0, mask = 1; i < 8; i++, mask <<= 1 )
      if( get_set_keys.released & mask )
        printf( " %d", i + 1 );
    printf( "\n" );
  }
}

//get current mode and default mode
static void cmd_mode_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_mode get_mode;
  int err;

  //get mode and default mode
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_mode, &get_mode );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show mode and default mode
  else
  {
    //show mode
    switch( get_mode.mode )
    {
      case am_usb_mode_null: printf( "  mode: null\n" ); break;
      case am_usb_mode_stream: printf( "  mode: stream\n" ); break;
      case am_usb_mode_cf: printf( "  mode: cf\n" ); break;
      default: printf( "  mode: <unknown>\n" );
    }
    //show default mode
    switch( get_mode.def_mode )
    {
      case am_usb_mode_null: printf( "  default mode: null\n" ); break;
      case am_usb_mode_stream: printf( "  default mode: stream\n" ); break;
      case am_usb_mode_cf: printf( "  default mode: cf\n" ); break;
      default: printf( "  default mode: <unknown>\n" );
    }
  }
}

//set current mode and default mode
static void cmd_mode_set( int fd_am_usb, char * p_params )
{
  char * p_mode, * p_def_mode;
  struct t_am_usb_set_mode set_mode;
  int err;

  //get mode
  str_first_word( p_params, &p_mode, &p_params );
  if( p_mode[0] == 0 )
    set_mode.mode = am_usb_mode_none;
  else if( strcasecmp( p_mode, "null" ) == 0 )
    set_mode.mode = am_usb_mode_null;
  else if( strcasecmp( p_mode, "stream" ) == 0 )
    set_mode.mode = am_usb_mode_stream;
  else if( strcasecmp( p_mode, "cf" ) == 0 )
    set_mode.mode = am_usb_mode_cf;
  else
  {
    printf( "  invalid mode: \"%s\"\n", p_mode );
    return;
  }
  
  //get default mode
  str_first_word( p_params, &p_def_mode, &p_params );
  if( p_def_mode[0] == 0 )
    set_mode.def_mode = am_usb_mode_none;
  else if( strcasecmp( p_def_mode, "null" ) == 0 )
    set_mode.def_mode = am_usb_mode_null;
  else if( strcasecmp( p_def_mode, "stream" ) == 0 )
    set_mode.def_mode = am_usb_mode_stream;
  else if( strcasecmp( p_def_mode, "cf" ) == 0 )
    set_mode.def_mode = am_usb_mode_cf;
  else
  {
    printf( "  invalid mode: \"%s\"\n", p_def_mode );
    return;
  }

  //get mode and default mode
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_mode, &set_mode );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show mode and default mode
  else
  {
    //show mode
    switch( set_mode.mode )
    {
      case am_usb_mode_null: printf( "  mode: null\n" ); break;
      case am_usb_mode_stream: printf( "  mode: stream\n" ); break;
      case am_usb_mode_cf: printf( "  mode: cf\n" ); break;
      default: printf( "  mode: <unknown>\n" );
    }
    //show default mode
    switch( set_mode.def_mode )
    {
      case am_usb_mode_null: printf( "  default mode: null\n" ); break;
      case am_usb_mode_stream: printf( "  default mode: stream\n" ); break;
      case am_usb_mode_cf: printf( "  default mode: cf\n" ); break;
      default: printf( "  default mode: <unknown>\n" );
    }
  }
}

//get output greyscale mapping
static void cmd_out_greymap_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_out_greymap get_out_greymap;
  int err, i;

  //get output greyscale mapping
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_out_greymap, &get_out_greymap );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show output greyscale mapping
  else
  {
    //success
    if( get_out_greymap.err == 0 )
    {
      printf( "  grayscale:  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15\n" );
      printf( "  mapped to:" );
      for( i = 0; i < sizeof( get_out_greymap.greymap ); i++ )
        printf( " %2d", get_out_greymap.greymap[i] );
      printf( "\n" );
    }
    //error
    else
      printf( "  reading output greyscale mapping failed (maybe EEPROM is busy)\n" );
  }
}

//set output greyscale mapping
static void cmd_out_greymap_set( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long value;
  struct t_am_usb_set_out_greymap set_out_greymap;
  int i, err;

  //get output greyscale mapping from parameters
  for( i = 0; i < count( set_out_greymap.greymap ); i++ )
  {
    str_first_word( p_params, &p_param, &p_params );
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value < 1 || value > 15 )
    {
      printf( "  invalid greyscale mapping for grayscale %d: \"%s\"\n", i + 1, p_param );
      return;
    }
    set_out_greymap.greymap[i] = (unsigned char)value;
  }

  //set output greyscale mapping
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_out_greymap, &set_out_greymap );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show result
  else
  {
    //error
    if( set_out_greymap.err != 0 )
      printf( "  writing output greyscale mapping failed (maybe EEPROM is busy)\n" );
  }
}

//get output style
static void cmd_out_style_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_out_style get_out_style;
  int err;

  //get output style
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_out_style, &get_out_style );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show output style
  else
  {
    //success
    if( get_out_style.err == 0 )
    {
      printf( "  output style: " );
      switch( get_out_style.style )
      {
        case am_usb_out_style_led: printf( "LED" ); break;
        case am_usb_out_style_latch: printf( "LATCH" ); break;
        case am_usb_out_style_amp: printf( "AMP" ); break;
        default: printf( "unknown (0x%02X)", get_out_style.style );
      };
      printf( "\n" );
    }
    //error
    else
      printf( "  reading output style failed (maybe EEPROM is busy)\n" );
  }
}

//set output style
static void cmd_out_style_set( int fd_am_usb, char * p_params )
{
  char * p_param;
  struct t_am_usb_set_out_style set_out_style;
  int err;

  //get output style from parameter
  str_first_word( p_params, &p_param, &p_params );
  if( strcasecmp( p_param, "led" ) == 0 )
    set_out_style.style = am_usb_out_style_led;
  else if( strcasecmp( p_param, "latch" ) == 0 )
    set_out_style.style = am_usb_out_style_latch;
  else if( strcasecmp( p_param, "amp" ) == 0 )
    set_out_style.style = am_usb_out_style_amp;
  else
  {
    printf( "  invalid output style: \"%s\"\n", p_param );
    return;
  }

  //set output style
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_out_style, &set_out_style );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show result
  else
  {
    //error
    if( set_out_style.err != 0 )
      printf( "  writing output style failed (maybe EEPROM is busy)\n" );
  }
}

//get output timing
static void cmd_out_timing_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_out_timing get_out_timing;
  int err, i;

  //get output timing
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_out_timing, &get_out_timing );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show output timing
  else
  {
    //success
    if( get_out_timing.err == 0 )
    {
      printf( "  grayscale:      1  2  3  4  5  6  7  8  9 10 11 12 13 14 15\n" );
      printf( "  output timing:" );
      for( i = 0; i < sizeof( get_out_timing.timing ); i++ )
        printf( " %2d", get_out_timing.timing[i] );
      printf( "\n" );
    }
    //error
    else
      printf( "  reading output timing failed (maybe EEPROM is busy)\n" );
  }
}

//set output timing
static void cmd_out_timing_set( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long value;
  struct t_am_usb_set_out_timing set_out_timing;
  int i, err;

  //get output timing from parameters
  for( i = 0; i < count( set_out_timing.timing ); i++ )
  {
    str_first_word( p_params, &p_param, &p_params );
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 50 )
    {
      printf( "  invalid timing for grayscale %d: \"%s\"\n", i + 1, p_param );
      return;
    }
    set_out_timing.timing[i] = (unsigned char)value;
  }

  //set output timing
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_out_timing, &set_out_timing );

  //show error
  if( err != 0 )
    printf( "  error: %s\n", strerror( errno ) );
  //show result
  else
  {
    //error
    if( set_out_timing.err != 0 )
      printf( "  writing output timing failed (maybe EEPROM is busy)\n" );
  }
}

//load (a part of) the overlay pixel buffer
static void cmd_ovp_buf_load( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long start, value;
  struct t_am_usb_read_ovp_buf read_ovp_buf;
  struct t_am_usb_load_ovp_buf load_ovp_buf;
  int err, i, j;

  //get start address
  str_first_word( p_params, &p_param, &p_params );
  start = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || start > 259 )
  {
    printf( "  invalid start address: \"%s\"\n", p_param );
    return;
  }

  //read overlay pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_ovp_buf, &read_ovp_buf );

  //show error
  if( err != 0 )
  {
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
    return;
  }

  //modify data
  memcpy( load_ovp_buf.pixels, read_ovp_buf.pixels, sizeof( read_ovp_buf.pixels ) );
  for( i = (int)start; i < 260; )
  {
    //get next data byte
    str_first_word( p_params, &p_param, &p_params );
    if( p_param[0] == 0 )
      break;
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 255 )
      printf( "  ignoring invalid data byte: \"%s\"\n", p_param );
    else
      load_ovp_buf.pixels[i++] = (unsigned char)value;
  }

  //load overlay pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_load_ovp_buf, &load_ovp_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show overlay pixel buffer data
  else
  {
    for( i = 0; i < 20; i++ )
    {
      printf( "  %03d:", i * 13 );
      for( j = 0; j < 13; j++ )
        printf( " %02X", load_ovp_buf.pixels[i * 13 + j] );
      printf( "\n" );
    }
  }
}

//read the overlay pixel buffer
static void cmd_ovp_buf_read( int fd_am_usb, char * p_params )
{
  struct t_am_usb_read_ovp_buf read_ovp_buf;
  int err, i, j;

  //read overlay pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_ovp_buf, &read_ovp_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show overlay pixel buffer data
  else
  {
    for( i = 0; i < 20; i++ )
    {
      printf( "  %03d:", i * 13 );
      for( j = 0; j < 13; j++ )
        printf( " %02X", read_ovp_buf.pixels[i * 13 + j] );
      printf( "\n" );
    }
  }
}

//load (a part of) the pixel buffer
static void cmd_pix_buf_load( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long start, value;
  struct t_am_usb_read_pix_buf read_pix_buf;
  struct t_am_usb_load_pix_buf load_pix_buf;
  int err, i, j;

  //get start address
  str_first_word( p_params, &p_param, &p_params );
  start = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || start > 259 )
  {
    printf( "  invalid start address: \"%s\"\n", p_param );
    return;
  }

  //read pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_pix_buf, &read_pix_buf );

  //show error
  if( err != 0 )
  {
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
    return;
  }

  //modify data
  memcpy( load_pix_buf.pixels, read_pix_buf.pixels, sizeof( read_pix_buf.pixels ) );
  for( i = (int)start; i < 260; )
  {
    //get next data byte
    str_first_word( p_params, &p_param, &p_params );
    if( p_param[0] == 0 )
      break;
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 255 )
      printf( "  ignoring invalid data byte: \"%s\"\n", p_param );
    else
      load_pix_buf.pixels[i++] = (unsigned char)value;
  }

  //load pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_load_pix_buf, &load_pix_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show pixel buffer data
  else
  {
    for( i = 0; i < 20; i++ )
    {
      printf( "  %03d:", i * 13 );
      for( j = 0; j < 13; j++ )
        printf( " %02X", load_pix_buf.pixels[i * 13 + j] );
      printf( "\n" );
    }
  }
}

//read the pixel buffer
static void cmd_pix_buf_read( int fd_am_usb, char * p_params )
{
  struct t_am_usb_read_pix_buf read_pix_buf;
  int err, i, j;

  //read pixel buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_pix_buf, &read_pix_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show pixel buffer data
  else
  {
    for( i = 0; i < 20; i++ )
    {
      printf( "  %03d:", i * 13 );
      for( j = 0; j < 13; j++ )
        printf( " %02X", read_pix_buf.pixels[i * 13 + j] );
      printf( "\n" );
    }
  }
}

//run ARCADEmini
static void cmd_run( int fd_am_usb, char * p_params )
{
  int err;

  //run ARCADEmini
  err = ioctl( fd_am_usb, am_usb_ioctl_no_run, NULL );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
}

//load (a part of) the sector buffer
static void cmd_sec_buf_load( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long start, value;
  struct t_am_usb_read_sec_buf read_sec_buf;
  struct t_am_usb_load_sec_buf load_sec_buf;
  int err, i, j;

  //get start address
  str_first_word( p_params, &p_param, &p_params );
  start = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || start > 511 )
  {
    printf( "  invalid start address: \"%s\"\n", p_param );
    return;
  }

  //read sector buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_sec_buf, &read_sec_buf );

  //show error
  if( err != 0 )
  {
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
    return;
  }

  //modify data
  memcpy( load_sec_buf.data, read_sec_buf.data, sizeof( read_sec_buf.data ) );
  for( i = (int)start; i < 512; )
  {
    //get next data byte
    str_first_word( p_params, &p_param, &p_params );
    if( p_param[0] == 0 )
      break;
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 255 )
      printf( "  ignoring invalid data byte: \"%s\"\n", p_param );
    else
      load_sec_buf.data[i++] = (unsigned char)value;
  }

  //load sector buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_load_sec_buf, &load_sec_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector buffer data
  else
  {
    for( i = 0; i < 32; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", load_sec_buf.data[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//read the sector buffer
static void cmd_sec_buf_read( int fd_am_usb, char * p_params )
{
  struct t_am_usb_read_sec_buf read_sec_buf;
  int err, i, j;

  //read sector buffer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_sec_buf, &read_sec_buf );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector buffer data
  else
  {
    for( i = 0; i < 32; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", read_sec_buf.data[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//get key SNES0 state (and clear it)
static void cmd_snes0( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_set_snes0 get_set_snes0;
  int err, i;
  unsigned short mask;

  //clear key state
  get_set_snes0.in_pressed = 0x0000;
  get_set_snes0.in_pressed_mask = 0xFFFF;
  get_set_snes0.in_released = 0x0000;
  get_set_snes0.in_released_mask = 0xFFFF;

  //get and set key state
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_set_snes0, &get_set_snes0 );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show key state
  else
  {
    printf( "  state:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes0.state & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
    printf( "  pressed:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes0.pressed & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
    printf( "  released:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes0.released & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
  }
}

//get key SNES1 state (and clear it)
static void cmd_snes1( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_set_snes1 get_set_snes1;
  int err, i;
  unsigned short mask;

  //clear key state
  get_set_snes1.in_pressed = 0x0000;
  get_set_snes1.in_pressed_mask = 0xFFFF;
  get_set_snes1.in_released = 0x0000;
  get_set_snes1.in_released_mask = 0xFFFF;

  //get and set key state
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_set_snes1, &get_set_snes1 );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show key state
  else
  {
    printf( "  state:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes1.state & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
    printf( "  pressed:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes1.pressed & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
    printf( "  released:" );
    for( i = 0, mask = 1; i < 16; i++, mask <<= 1 )
      if( get_set_snes1.released & mask )
        printf( " %s", SNES_KEYS[i] );
    printf( "\n" );
  }
}

//load (a part of) the stack
static void cmd_stack_load( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long start, value;
  struct t_am_usb_read_stack read_stack;
  struct t_am_usb_load_stack load_stack;
  int err, i, j;

  //get start address
  str_first_word( p_params, &p_param, &p_params );
  start = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || start > 127 )
  {
    printf( "  invalid start address: \"%s\"\n", p_param );
    return;
  }

  //read stack
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_stack, &read_stack );

  //show error
  if( err != 0 )
  {
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
    return;
  }

  //modify data
  memcpy( load_stack.data, read_stack.data, sizeof( read_stack.data ) );
  for( i = (int)start; i < 128; )
  {
    //get next data byte
    str_first_word( p_params, &p_param, &p_params );
    if( p_param[0] == 0 )
      break;
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 255 )
      printf( "  ignoring invalid data byte: \"%s\"\n", p_param );
    else
      load_stack.data[i++] = (unsigned char)value;
  }

  //load stack
  err = ioctl( fd_am_usb, am_usb_ioctl_no_load_stack, &load_stack );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector buffer data
  else
  {
    for( i = 0; i < 8; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", load_stack.data[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//read the stack
static void cmd_stack_read( int fd_am_usb, char * p_params )
{
  struct t_am_usb_read_stack read_stack;
  int err, i, j;

  //read stack
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_stack, &read_stack );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show sector buffer data
  else
  {
    for( i = 0; i < 8; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", read_stack.data[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//get stack pointer
static void cmd_stack_ptr_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_stack_ptr get_stack_ptr;
  int err;

  //get stack pointer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_stack_ptr, &get_stack_ptr );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show command pointer
  else
    printf( "  stack pointer: %X\n",
            get_stack_ptr.stack_ptr );
}

//set stack pointer
static void cmd_stack_ptr_set( int fd_am_usb, char * p_params )
{
  struct t_am_usb_set_stack_ptr set_stack_ptr;
  char * p_param, * p_chr;
  unsigned long val;
  int err;

  //get stack pointer
  str_first_word( p_params, &p_param, &p_params );
  val = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || val > 127 )
  {
    printf( "  invalid stack pointer value: \"%s\"\n", p_param );
    return;
  }
  set_stack_ptr.stack_ptr = (unsigned char)val;

  //set stack pointer
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_stack_ptr, &set_stack_ptr );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show command pointer
  else
    printf( "  stack pointer: %X\n",
            set_stack_ptr.stack_ptr );
}

//step ARCADEmini
static void cmd_step( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long cnt;
  int i, err;

  //get number of steps
  str_first_word( p_params, &p_param, &p_params );
  cnt = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 )
    cnt = 1;
  else
  { 
    if( p_chr[0] != 0 || cnt > 1000 )
    {
      printf( "  invalid step count: \"%s\"\n", p_param );
      return;
    }
  }

  //step ARCADEmini cnt times
  for( i = 0, err = 0; i < (int)cnt && err == 0; i++ )
    err = ioctl( fd_am_usb, am_usb_ioctl_no_step, NULL );

  //show error
  if( err != 0 )
    printf( "  error (after %d steps): %s (%d)\n", i, strerror( errno ), errno );
}

//swap pixel buffers
static void cmd_swap_bufs( int fd_am_usb, char * p_params )
{
  int err;

  //swap pixel buffers
  err = ioctl( fd_am_usb, am_usb_ioctl_no_swap_bufs, NULL );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
}

//swap pixel buffers and copy foreground to background buffer
static void cmd_swap_bufs_cp( int fd_am_usb, char * p_params )
{
  int err;

  //swap pixel buffers and copy image
  err = ioctl( fd_am_usb, am_usb_ioctl_no_swap_bufs_cp, NULL );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
}

//get time from the ARCADEmini clock
static void cmd_time_get( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_time get_time;
  int err;

  //get time
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_time, &get_time );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show time
  else
    printf( "  time is %02d:%02d:%02d\n",
            get_time.hour, get_time.min, get_time.sec );
}

//set time of the ARCADEmini clock
static void cmd_time_set( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long val;
  time_t gmttime;
  struct tm * p_localtime;
  struct t_am_usb_set_time set_time;
  int err;

  //time specified
  str_first_word( p_params, &p_param, &p_params );
  if( p_param[0] != 0 )
  {
    //get hour
    val = strtoul( p_param, &p_chr, 0 );
    if( p_chr[0] != 0 || val > 23 )
    {
      printf( "  invalid hour: \"%s\"\n", p_param );
      return;
    }
    set_time.hour = (unsigned char)val;
    //get minute
    str_first_word( p_params, &p_param, &p_params );
    val = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || val > 59 )
    {
      printf( "  invalid minute: \"%s\"\n", p_param );
      return;
    }
    set_time.min = (unsigned char)val;
    //get second
    str_first_word( p_params, &p_param, &p_params );
    val = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || val > 59 )
    {
      printf( "  invalid second: \"%s\"\n", p_param );
      return;
    }
    set_time.sec = (unsigned char)val;
  }
  //time not specified
  else
  {
    //get local time from system
    gmttime = time( NULL );
    p_localtime = localtime( &gmttime );
    set_time.hour = p_localtime->tm_hour;
    set_time.min = p_localtime->tm_min;
    set_time.sec = p_localtime->tm_sec;
  }

  //set time
  err = ioctl( fd_am_usb, am_usb_ioctl_no_set_time, &set_time );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
}

//load (a part of) the variables
static void cmd_vars_load( int fd_am_usb, char * p_params )
{
  char * p_param, * p_chr;
  unsigned long start, value;
  struct t_am_usb_read_vars read_vars;
  struct t_am_usb_load_vars load_vars;
  int err, i, j;

  //get start address
  str_first_word( p_params, &p_param, &p_params );
  start = strtoul( p_param, &p_chr, 0 );
  if( p_param[0] == 0 || p_chr[0] != 0 || start > 255 )
  {
    printf( "  invalid start address: \"%s\"\n", p_param );
    return;
  }

  //read variables
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_vars, &read_vars );

  //show error
  if( err != 0 )
  {
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
    return;
  }

  //modify data
  memcpy( load_vars.vars, read_vars.vars, sizeof( read_vars.vars ) );
  for( i = (int)start; i < 256; )
  {
    //get next data byte
    str_first_word( p_params, &p_param, &p_params );
    if( p_param[0] == 0 )
      break;
    value = strtoul( p_param, &p_chr, 0 );
    if( p_param[0] == 0 || p_chr[0] != 0 || value > 255 )
      printf( "  ignoring invalid data byte: \"%s\"\n", p_param );
    else
      load_vars.vars[i++] = (unsigned char)value;
  }

  //load variables
  err = ioctl( fd_am_usb, am_usb_ioctl_no_load_vars, &load_vars );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show variables
  else
  {
    for( i = 0; i < 16; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", load_vars.vars[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//read the variables
static void cmd_vars_read( int fd_am_usb, char * p_params )
{
  struct t_am_usb_read_vars read_vars;
  int err, i, j;

  //read variables
  err = ioctl( fd_am_usb, am_usb_ioctl_no_read_vars, &read_vars );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show variables
  else
  {
    for( i = 0; i < 16; i++ )
    {
      printf( "  %03X:", i << 4 );
      for( j = 0; j < 16; j++ )
        printf( " %02X", read_vars.vars[(i << 4) + j] );
      printf( "\n" );
    }
  }
}

//get version of driver
static void cmd_ver( int fd_am_usb, char * p_params )
{
  struct t_am_usb_get_version get_version;
  int err;

  //get driver version
  err = ioctl( fd_am_usb, am_usb_ioctl_no_get_version, &get_version );

  //show error
  if( err != 0 )
    printf( "  error: %s (%d)\n", strerror( errno ), errno );
  //show driver version
  else
    printf( "  version %d.%d.%d, date %04d-%02d-%02d\n",
            get_version.major, get_version.minor, get_version.revision,
            get_version.year, get_version.month, get_version.day );
}

//help command prototype
static void cmd_help( int fd_am_usb, char * p_params );

//command table
struct t_cmd_tab
{
  char * name;
  char * parameters;
  char * description;
  void (* function)( int fd_am_usb, char * p_params );
} cmd_tab[] =
{
  { "cf-flag", "", "check if commands are executed from compact flash or internal flash", cmd_cf_flag },
  { "cf-read", " <sector_no>", "read compact flash sector into sector buffer", cmd_cf_read },
  { "cf-secs", "", "get number of sectors on compact flash", cmd_cf_secs },
  { "cf-write", " <sector_no>", "write sector buffer to compact flash sector", cmd_cf_write },
  { "cmd-ptr-get", "", "get command pointer", cmd_cmd_ptr_get },
  { "cmd-ptr-set", " <sector_no> <offset>", "set command pointer", cmd_cmd_ptr_set },
  { "end", "", "end the program", NULL },
  { "exit", "", "end the program", NULL },
  { "fw-ver", "", "show firmware version", cmd_fw_ver },
  { "halt", "", "halt ARCADEmini", cmd_halt },
  { "help", "", "you just found out what this does :-)", cmd_help },
  { "id-get", "", "get ID of this ARCADEmini", cmd_id_get },
  { "id-set", " <id>", "set ID of this ARCADEmini", cmd_id_set },
  { "if-read", " <sector_no>", "read internal flash sector into sector buffer", cmd_if_read },
  { "if-secs", "", "get number of sectors in internal flash", cmd_if_secs },
  { "keys", "", "get key state (and clear it)", cmd_keys },
  { "mode-get", "", "get current mode and default mode", cmd_mode_get },
  { "mode-set", " [<mode> [<default mode>]]", "set current mode and default mode", cmd_mode_set },
  { "out-greymap-get", "", "get output greyscale mapping", cmd_out_greymap_get },
  { "out-greymap-set", " <mapping 1> ... <mapping 15>", "set output greyscale mapping", cmd_out_greymap_set },
  { "out-style-get", "", "get output style", cmd_out_style_get },
  { "out-style-set", " [led|latch|amp]", "set output style", cmd_out_style_set },
  { "out-timing-get", "", "get output timing", cmd_out_timing_get },
  { "out-timing-set", " <timing 1> ... <timing 15>", "set output timing", cmd_out_timing_set },
  { "ovp-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the overlay pixel buffer", cmd_ovp_buf_load },
  { "ovp-buf-read", "", "read the overlay pixel buffer", cmd_ovp_buf_read },
  { "pix-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the pixel buffer", cmd_pix_buf_load },
  { "pix-buf-read", "", "read the pixel buffer", cmd_pix_buf_read },
  { "quit", "", "end the program", NULL },
  { "run", "", "run ARCADEmini", cmd_run },
  { "sec-buf-load", " <start-addr> [<data> [...]]", "load (a part of) the sector buffer", cmd_sec_buf_load },
  { "sec-buf-read", "", "read the sector buffer", cmd_sec_buf_read },
  { "snes0", "", "get key SNES0 state (and clear it)", cmd_snes0 },
  { "snes1", "", "get key SNES1 state (and clear it)", cmd_snes1 },
  { "stack-load", " <start-addr> [<data> [...]]", "load (a part of) the stack", cmd_stack_load },
  { "stack-read", "", "read the stack", cmd_stack_read },
  { "stack-ptr-get", "", "get stack pointer", cmd_stack_ptr_get },
  { "stack-ptr-set", " <stack_ptr>", "set stack pointer", cmd_stack_ptr_set },
  { "step", " [<count>]", "step ARCADEmini", cmd_step },
  { "swap-bufs", "", "swap pixel buffers", cmd_swap_bufs },
  { "swap-bufs-cp", "", "swap pixel buffers and copy foreground to background buffer", cmd_swap_bufs_cp },
  { "time-get", "", "get time from the ARCADEmini clock", cmd_time_get },
  { "time-set", " [<hour> <min> <sec>]", "set time of the ARCADEmini clock", cmd_time_set },
  { "vars-load", " <start-addr> [<data> [...]]", "load (a part of) the variables", cmd_vars_load },
  { "vars-read", "", "read the variables", cmd_vars_read },
  { "ver", "", "show driver version", cmd_ver },
};

//help command
static void cmd_help( int fd_am_usb, char * p_params )
{
  char * p_cmd;
  int i;

  //get command
  str_first_word( p_params, &p_cmd, &p_params );

  //command found in parameters
  if( p_cmd[0] != 0 )
  {
    //search command
    for( i = 0; i < count( cmd_tab ); i++ )
      if( strcasecmp( p_cmd, cmd_tab[i].name ) == 0 )
        break;
    //command found in table
    if( i < count( cmd_tab ) )
      //print help for this command
      printf( "  %s%s\n    %s\n", cmd_tab[i].name, cmd_tab[i].parameters, cmd_tab[i].description );
    //command not found in table
    else
      //print error message
      printf( "  no help found to \"%s\"\n", p_cmd );
    return;
  }

  //print command table and descriptions
  for( i = 0; i < count( cmd_tab ); i++ )
    printf( "  %s%s\n    %s\n", cmd_tab[i].name, cmd_tab[i].parameters, cmd_tab[i].description );
}

//main program
int main( int arg_cnt, char * * args )
{
  int fd_am_usb, i;
  char line[1024], * p_chr, * p_cmd, * p_params;

  //print message
  printf( "ARCADEmini USB driver for Linux test application\n"
          "version 1.2.0 date 2005-06-06\n"
          "part of the ARCADEmini project - http[s]://arcademini.schuermans.info/\n"
          "Copyright (C) 2003-2004 1stein <1stein@schuermans.info>\n"
          "Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html\n" );

  //check parameters
  if( arg_cnt < 2 )
  {
    printf( "  usage: %s <am_usb device>\n", args[0] );
    return -1;
  }

  //open am_usb device
  fd_am_usb = open( args[1], O_RDWR );
  if( fd_am_usb == -1 )
  {
    printf( "  could not open am_usb device \"%s\": %s\n", args[1], strerror( errno ) );
    return -1;
  }

  //main loop
  for( ; ; )
  {
    //read a line
    printf( "am_usb> " ); //prompt
    fflush( stdout );
    line[0] = 0; //read line
    fgets( line, sizeof( line ), stdin );
    p_chr = strchr( line, '\n' ); //remove newline
    if( p_chr != NULL )
      *p_chr = 0;

    //get command (first word of line)
    str_first_word( line, &p_cmd, &p_params );

    //ignore empty command
    if( p_cmd[0] != 0 )
    {
      //find command in command table
      for( i = 0; i < count( cmd_tab ); i++ )
        if( strcasecmp( p_cmd, cmd_tab[i].name ) == 0 )
          break;
      //command found
      if( i < count( cmd_tab ) )
      {
        //exit main loop if no function
        if( cmd_tab[i].function == NULL )
          break;
        //call command
        cmd_tab[i].function( fd_am_usb, p_params );
      }
      //command not found
      else
        printf( "  unknown command - type \"help\"\n" );
    }
  }

  //close am_usb device
  close( fd_am_usb );

  return 0;
}

